"""
    author: Sean Kullmann, Hosung Lee
    date: December 7 2020

    AES class
"""
from abc import abstractmethod
from typing import Optional, List
from Cryptodome.Cipher import AES
from Cryptodome.Cipher.AES import MODE_CCM, MODE_EAX, MODE_GCM, MODE_SIV, MODE_OCB
from photocrypt.core import Cipher, packer


class AESCipher(Cipher):
    """
    Abstract AES Cipher class
    """

    def __init__(self, mode, key: bytes):
        self._mode = mode
        self._key = key

    @property
    def mode(self):
        """
        getter method of mode
        """
        return self._mode

    @property
    def key(self):
        """
        getter method of key
        """
        return self._key

    @abstractmethod
    def encrypt(self, data: bytes, extra: Optional[bytes] = None) -> List[bytes]:
        """
        encrypts bytes

            Parameters:
                data (bytes): data to encrypt using AES
                extra (Optional[bytes]): optional data required for encryption

            Return:
                encrypted (bytes): encrypted data
                extra (bytes): extra data produced from decryption
        """
        ...

    @abstractmethod
    def decrypt(self, data: bytes, extra: Optional[bytes] = None) -> List[bytes]:
        """
        decrypts bytes

            Parameters:
                data (bytes): data to decrypt using AES
                extra (Optional[bytes]): packed bytes for extra
                                            (packed using photocrypt.packer).

            Return:
                decrypted (bytes): encrypted data
                extra (Optional[bytes]): packed bytes for extra
                                            (packed using photocrypt.packer).
        """
        ...


class AESCipherModern(AESCipher):
    """
    AES Cipher class
    """

    def __init__(self, mode: int, key: bytes, extra: Optional[bytes] = None):
        extra = packer.unpack(extra) if extra else []
        _nonce = None
        if len(extra) > 0:
            _nonce = extra[0]
        self.aes = AES.new(key, mode, nonce=_nonce)
        self._nonce = _nonce
        super().__init__(mode, key)

    def encrypt(self, data: bytes, extra: Optional[bytes] = None) -> List[bytes]:
        """
        encrypts bytes

            Parameters:
                data (bytes): data to encrypt using AES
                extra (Optional[bytes]): packed bytes for extra
                                            (packed using photocrypt.packer).

            Return:
                encrypted (bytes): encrypted data
                extra (bytes): packed bytes of nonce and tag
                                            (packed using photocrypt.packer).
        """

        nonce = self._nonce if self.mode == MODE_SIV else self.aes.nonce
        ciphertext, tag = self.aes.encrypt_and_digest(data)
        return ciphertext, packer.pack(nonce, tag)

    def decrypt(self, data: bytes, extra: Optional[bytes] = None) -> List[bytes]:
        """
        decrypts bytes

            Parameters:
                data (bytes): data to decrypt using AES
                extra (Optional[bytes]): packed bytes for extra
                                            (packed using photocrypt.packer).

            Return:
                decrypted (bytes): encrypted data
                extra: packed bytes for extra
                                            (packed using photocrypt.packer).
        """
        tag = packer.unpack(extra)[0]
        return self.aes.decrypt_and_verify(data, tag), packer.pack(b'')


# supported modes
_supported_modes = {
    MODE_CCM: AESCipherModern,
    MODE_EAX: AESCipherModern,
    MODE_GCM: AESCipherModern,
    MODE_SIV: AESCipherModern,
    MODE_OCB: AESCipherModern
}


def _get_aes(mode, key, extra: Optional[bytes] = None) -> AESCipher:
    """
    Factory of AES Ciphers

        Parameters:
            mode (constant): mode of AES
            key (bytes): key for AES cipher
            extra (bytes): packed bytes for extra information
                (packed using photocrypt.packer).
        
        Returns:
            AES cipher (AESCipher): AESCipher object.
    """
    if not mode in _supported_modes:
        raise ValueError("The mode is not supported.")

    return _supported_modes[mode](mode, key, extra)


def create(mode, key, extra: Optional[bytes] = None) -> AESCipher:
    """
    Creates AES Cipher

        Parameters:
            mode (constant): mode of AES
            key (bytes): key for cipher
            extra (bytes): packed bytes for extra information
                (packed using photocrypt.packer).
        
        Returns:
            AES cipher (AESCipher): AESCipher object.
    """
    return _get_aes(mode, key, extra)



"""
    author: Hosung Lee
    date: December 7 2020

    Abstract iamge and image header classes that are base classes of
    all Images and their headers within photocrypt package.
"""

from abc import abstractmethod
from typing import Union, List
import PIL.Image
from .bstream import ByteStream
from .bdata import ByteData


class ImageHeader(ByteData):
    """
    base class to represent Image Header
    """
    protocol = []

    def __init__(self, header: list):
        self.protocol = self.__class__.protocol
        self.header = header

    @classmethod
    def get_protocol(cls) -> List[type]:
        """
        get protocol of the header

            Return:
                protocol (list of types): protocol of current header.
        """
        return cls.protocol

    def get_value(self, i: int) -> Union[bytes, 'short', int, str]:
        """
        get header list

            Parameters:
                i (int): index of header
            
            Returns: 
                value (any): value in ith index of header.
        """
        if i < 0 or i > len(self.header):
            raise IndexError("header index out of bounds")
        return self.header[i]

    def set_value(self, i: int, value: Union[bytes, 'short', int, str]) -> None:
        """
        sets header value

            Parameters:
                i (int): index of header
                value (any) value to set in ith index of header.
        """
        if i < 0 or i > len(self.header):
            raise IndexError("header index out of bounds")
        self.header[i] = value

    @classmethod
    def read(cls, data_stream: ByteStream) -> 'ImageHeader':
        """
        Reads ByteStream to generate a header.

            Parameters:
                data_stream (ByteStream): bytestream to read
            
            Return:
                image header (ImageHeader): image header object
        """

        return cls(data_stream.read_multiple(cls.get_protocol()))

    def write(self, data_stream: ByteStream) -> None:
        """
        Write header to a ByteStream

            Parameters:
                data_stream (ByteStream): bytestream to write
        """
        data_stream.write_multiple([
            (x[0], y) if isinstance(x, tuple) else (x, y)
            for x, y
            in zip(self.protocol, self.header)
        ])

    def __repr__(self):
        return f"<{self.__class__.__name__}>[{' | '.join(map(str,self.header))}]"


class Image(ByteData):
    """
    abstract class to represent Image types
    """
    image_format = None

    def __init__(self, headers: list, data: bytes):
        self._headers = headers
        self._data = data

    @classmethod
    @abstractmethod
    def test_format(cls, data: bytes) -> Union[str, None]:
        """
        returns name of format if data represents current format, None otherwise.

            Parameters:
                data (bytes): data that represents image.
            Returns:
                name of format (str) if data represents current format
                None otherwise.
        """
        ...

    @property
    def headers(self) -> List[ImageHeader]:
        """
        getter method of headers

            Returns:
                headers (list of image headers)
        """
        return self._headers

    @property
    def data(self) -> bytes:
        """
        getter method of data

            Returns:
                image data (bytes): data part of the image
        """
        return self._data

    @data.setter
    def data(self, data: bytes) -> None:
        """
        setter method of data

            Parameters:
                data (bytes): image data in bytes
        """
        if len(data) < len(self._data):
            raise ValueError(f"data is smaller than {len(self._data)}")

        if len(data) > len(self._data):
            self._data = data[:len(self._data)]

        else:
            self._data = data

    @classmethod
    @abstractmethod
    def read(cls, data_stream: ByteStream):
        """
        Read bytes from byte stream to construct the class.
        """
        ...

    @abstractmethod
    def write(self, data_stream: ByteStream) -> None:
        """
        write bytes to byte stream.
        """
        ...

    @classmethod
    def open(cls, file_path: str) -> 'Image':
        """
        loads image from path and converts it to specified format of class.
        """
        with open(file_path, 'rb') as file:
            head = file.read(32)
            if not cls.test_format(head):
                raise ValueError(f"File is not in a {cls.image_format}.")
            file.seek(0)
            return cls.from_bytes(file.read())

    def save(self, file_path: str) -> None:
        """
        saves image to path.
        """
        with open(file_path, 'wb') as file:
            file.write(self.to_bytes())


def open_image_as(file_path: str, as_format: Image) -> Image:
    """
    loads image from path and converts it to specified image format
    """
    try:
        return as_format.open(file_path)
    except ValueError:
        data_stream = ByteStream()
        PIL.Image.open(file_path).save(
            data_stream, format=as_format.image_format)
        data_stream.seek(0)
        return as_format.read(data_stream)


def convert_image(src: Image, to_format: Image) -> Image:
    """
    Converts an image from one type to another.
    """
    return to_format.from_bytes(src.to_bytes())


def _open_image(supported_formats: list, file_path: str) -> Image:
    """
    loads image from path to Image class.
    """
    if len(supported_formats) == 0:
        raise Exception("operation not supported yet.")
    _default_format = supported_formats[0]
    with open(file_path, 'rb') as file:
        head = file.read(32)
        for _f in supported_formats:
            if _f.test_format(head):
                return _f.open(file_path)

    return open_image_as(file_path, _default_format)



"""
Photocrypt Package
"""

# image operations
from .image import open_image
from .image import open_image_as

# byte operations
from .core import ByteData
from .core import ByteStream
from .core import packer

# encrypt and decrypt images
from .photocrypt import encrypt_image
from .photocrypt import decrypt_image



"""
    author: Hosung Lee
    date: December 7 2020

    abstract ByteData class that can be converted from/to bytes.
"""

from abc import ABC, abstractmethod
from .bstream import ByteStream

class ByteData(ABC):
    """
    abstract class to represent types that use bytes.
    """
    @classmethod
    @abstractmethod
    def read(cls, data_stream: ByteStream):
        """
        Read bytes from byte stream to construct the class.
        """
        ...

    @abstractmethod
    def write(self, data_stream: ByteStream) -> None:
        """
        write bytes to byte stream.
        """
        ...

    @classmethod
    def from_bytes(cls, data: bytes, offset: int = 0) -> 'ImageHeader':
        """
        Load header from bytes.
        """
        data_stream = ByteStream(data)
        data_stream.seek(offset)
        return cls.read(data_stream)

    def to_bytes(self) -> bytes:
        """
        Get bytes of header
        """
        data_stream = ByteStream()
        self.write(data_stream)
        return data_stream.getvalue()



"""
    author: Hosung Lee
    date: December 7 2020

    Local key manager to manage public keys
"""
import sqlite3
from sqlite3 import IntegrityError
from typing import List, Dict, Union

KEY_STORE_TABLE_NAME = 'keystore'
KEY_STORE_TABLE_SQL = f'''
    CREATE TABLE IF NOT EXISTS {KEY_STORE_TABLE_NAME} (
        ID INTEGER PRIMARY KEY AUTOINCREMENT,
        Name VARCHAR(50),
        Email VARCHAR(50),
        PublicKey BLOB UNIQUE
    )
'''

def conn_not_established():
    """
    Raise error that connection is not established.
    """
    raise Exception("connection not established. Call connect() method.")

def public_key_invalid():
    """
    Raise error that public key is invalid.
    """
    raise ValueError("public key is invalid.")

def public_key_used():
    """
    Raise error that public key is already in keystore.
    """
    raise ValueError("public key is already in keystore.")

def name_length_invalid():
    """
    Raise error that name is invalid.
    """
    raise ValueError("name has to be longer than 0.")

class KeyManager:
    """
    Key manager class
    """
    def __init__(self, dbpath=":memory:"):
        self.dbpath = dbpath
        self.conn = None

    def connect(self):
        """
        Connect to keystore
        """
        self.conn = sqlite3.connect(self.dbpath)
        cursor = self.conn.cursor()
        cursor.execute(KEY_STORE_TABLE_SQL)
        self.conn.commit()

    def list_key(self) -> List[Dict[str, Union[str, bytes]]]:
        """
        List all key in keystore
            Returns:
                list of dictionary[name, email, public_key]
        """
        if not self.conn:
            conn_not_established()

        cursor = self.conn.cursor()
        cursor.execute(f'SELECT * FROM {KEY_STORE_TABLE_NAME}')
        rows = cursor.fetchall()
        return [
            {'uid':i, 'name': name, 'email': email, 'public_key':public_key}
            for i, name, email, public_key
            in rows]

    def write_key(self, name: str, email: str, public_key: bytes) -> None:
        """
        Write key to keystore

            Parameters:
                name (str): required, name of key holder
                email (str): not required, email of key holder
                public_key (bytes): required, public_key of key holder
        """
        if not self.conn:
            conn_not_established()

        cursor = self.conn.cursor()
        if len(name) == 0:
            name_length_invalid()

        if len(public_key) == 0 or not isinstance(public_key, bytes):
            public_key_invalid()

        if not public_key.startswith(b"-----BEGIN PUBLIC KEY-----"):
            public_key_invalid()

        values = (name, email, public_key)

        try:
            cursor.execute(
                f'INSERT INTO {KEY_STORE_TABLE_NAME} (Name, Email, Publickey) VALUES(?, ?, ?)',
                values
                )

            self.conn.commit()

        except IntegrityError:
            public_key_used()

    def get_key(self, uid: int) -> bytes:
        """
        Get key by unique id

            Parameters:
                uid (int): required, unique id of key holder

            Returns
                public key (bytes): public key of key holder
        """
        if not self.conn:
            conn_not_established()

        cursor = self.conn.cursor()
        data = (uid, )

        cursor.execute(f'SELECT * FROM {KEY_STORE_TABLE_NAME} WHERE Id = ?', data)
        row = cursor.fetchone()
        if row:
            _, _, _, key = row
            return key

        return row

    def delete_key(self, uid: int) -> None:
        """
        Delete key by unique id

            Parameters:
                uid (int): required, unique id of key holder

            Returns
                public key (bytes): public key of key holder
        """
        if not self.conn:
            conn_not_established()

        cursor = self.conn.cursor()
        data = (uid, )

        cursor.execute(f'DELETE FROM {KEY_STORE_TABLE_NAME} WHERE Id = ?', data)
        self.conn.commit()

    def close(self):
        """
        Close connection to db
        """
        if self.conn:
            self.conn.close()

def create(keystore_path='keystore.db'):
    """
    Create key manager that controls specified keystore database.

        Parameters:
            keystore_path (str): path to keystore database.
                If there's no such file, key manager creates new one.

        Returns:
            key manager (KeyManager): key manager that controls keystore database.
    """
    keymgr = KeyManager(keystore_path)
    return keymgr



"""
    author: Hosung Lee
    date: December 7 2020

    Definition of Bitmap and Bitmap's header classes.
"""

from photocrypt.core import Image, ImageHeader, ByteStream

# data compression method
BI_RGB = 0
BI_RLE8 = 1
BI_RLE4 = 2
BI_BITFIELDS = 3
BI_JPEG = 4
BI_PNG = 5
BI_ALPHABITFIELDS = 6
BI_CMYK = 11
BI_CMYKRLE8 = 12
BI_CMYKRLE4 = 13

class BitmapFileHeader(ImageHeader):
    """
    class to represent bitmap file header.

    i   name            type
    =========================================
    0   signature       2 byte string
    1   file size       4 byte int
    2   reversed 1      2 byte short (unused)
    3   reversed 2      2 byte short (unused)
    4   data offset     4 byte int
    """

    protocol = [(str, [2]), int, 'short', 'short', int]

# index of bitmap file header
BF_SIGNATURE = 0
BF_FILE_SIZE = 1
BF_REVERSED_1 = 2
BF_REVERSED_2 = 3
BF_OFFSET = 4

class BitmapInfoHeader(ImageHeader):
    """
    class to represent bitmap information header.

    i   name            type
    =========================================
    0   size            4 byte int
    1   width           4 byte int
    2   height          4 byte int
    3   n_color_planes  2 byte short
    4   bpp             2 byte short
    5   compression     4 byte int
    6   image_size      4 byte int
    7   h_res           4 byte int
    8   v_res           4 byte int
    9   n_colors        4 byte int
    10  n_imp_colors    4 byte int
    """

    protocol = [*[int]*3, *['short']*2, *[int]*6]

# index of bitmap info header
BI_SIZE = 0
BI_WIDTH = 1
BI_HEIGHT = 2
BI_COLOR_PLANES = 3
BI_BPP = 4
BI_COMPRESSION = 5
BI_IMAGE_SIZE = 6
BI_HORIZ_RES = 7
BI_VERTI_RES = 8
BI_COLORS = 9
BI_IMP_COLORS = 10

class Bitmap(Image):
    """
    Image class to represent Bitmap
    """
    image_format = "BMP"

    def __init__(self, headers, data):
        super().__init__(headers, data)
        (
            self.file_header,
            self.info_header
        ) = headers[:2]

    @classmethod
    def test_format(cls, data: bytes):
        """
        returns name of format if format is valid, None otherwise.
        """
        if data.startswith(b'BM'):
            return cls.image_format
        return None

    @classmethod
    def read(cls, data_stream: ByteStream) -> 'Bitmap':
        """
        Load bitmap from bytes
        """
        file_header, info_header = cls.read_bitmap_headers(data_stream)

        data_stream.seek(file_header.get_value(BF_OFFSET))
        data = data_stream.read(info_header.get_value(BI_IMAGE_SIZE))

        new_bitmap = cls(
            [file_header, info_header],
            data
        )

        return new_bitmap

    def write(self, data_stream: ByteStream) -> None:
        """
        Get bytes of the bitmap.
        """

        for header in self.headers:
            header.write(data_stream)
        data_stream.write(self.data)

    @classmethod
    def read_bitmap_headers(cls, data_stream: ByteStream) -> (BitmapFileHeader, BitmapInfoHeader):
        """
        Reads bitmap headers from bytestream.
        """
        file_header = BitmapFileHeader.read(data_stream)
        info_header = BitmapInfoHeader.read(data_stream)
        image_size = info_header.get_value(BI_IMAGE_SIZE)
        if image_size == 0:
            header_length = len(file_header.header) + len(info_header.header)
            image_size = file_header.get_value(BF_FILE_SIZE) - header_length
            info_header.set_value(BI_IMAGE_SIZE, image_size)
        return file_header, info_header



"""
    author: Hosung Lee
    date: December 7 2020

    crypto bitmap class
"""

from photocrypt.core import ImageHeader, ByteStream, packer
from .bitmap import Bitmap, BitmapFileHeader, BF_FILE_SIZE, BF_OFFSET, BF_REVERSED_1, BI_IMAGE_SIZE

SIZE_INT = 4
CRYPTO_HEADER_INDICATOR = 4362
CH_LENGTH = 0
CH_DATA = 1

class CryptoHeader(ImageHeader):
    """
    Class to represent bitmap file header.

    i   name        type
    0   length      int (4 bytes)
    1   data        n bytes
    """

    protocol = [int, (bytes, [0])]

    def set_length(self, length: int):
        """
        Set length of data
        """
        self.protocol = [int, (bytes, [length])]

    def get_length(self):
        """
        get length of data in protocol
        """
        return self.protocol[1][1][0]


def create_crypto_header():
    """
    function to create CryptoHeader object
    """
    return CryptoHeader([0, b''])


class CryptoBitmap(Bitmap):
    """
    Extended Bitmap that includes a crypto header
    """
    def __init__(self, headers, data):
        super().__init__(headers, data)
        (
            self.crypto_header
        ) = headers[2]
    
    @classmethod
    def test_format(cls, data: bytes):
        """
        returns name of format if format is valid, None otherwise.
        """
        if data.startswith(b'BM') and data[6:8] == b'\n\x11':
            return cls.image_format
        return None

    def store_crypto_information(self, *data: bytes):
        """
        Store
        """
        data = packer.pack(*data)
        self.crypto_header.set_value(CH_DATA, data)
        self.crypto_header.set_value(CH_LENGTH, len(data))
        self.crypto_header.set_length(len(data))

    @classmethod
    def read(cls, data_stream: ByteStream) -> 'Bitmap':
        """
        Load bitmap from bytes
        """
        file_header, info_header = cls.read_bitmap_headers(data_stream)
        size = file_header.get_value(BF_FILE_SIZE)
        offset = file_header.get_value(BF_OFFSET)

        # using reversed_1 as indicator of cryptobitmap
        if file_header.get_value(BF_REVERSED_1) == CRYPTO_HEADER_INDICATOR:
            crypto_header = cls.read_crypto_header(data_stream)
            file_header.set_value(BF_FILE_SIZE, size - crypto_header.get_length() - SIZE_INT)
            file_header.set_value(BF_OFFSET, offset - crypto_header.get_length() - SIZE_INT)

        else:
            # if not a cryptobitmap, convert image to cryptobitmap
            crypto_header = create_crypto_header()
            file_header.set_value(BF_REVERSED_1, CRYPTO_HEADER_INDICATOR)
            offset += crypto_header.get_length()

        data_stream.seek(offset)
        data = data_stream.read(info_header.get_value(BI_IMAGE_SIZE))

        new_bitmap = cls(
            [file_header, info_header, crypto_header],
            data
        )

        return new_bitmap

    def write(self, data_stream: ByteStream) -> None:
        """
        Get bytes of the bitmap.
        """
        size = self.file_header.get_value(BF_FILE_SIZE)
        offset = self.file_header.get_value(BF_OFFSET)
        cdata_length = self.crypto_header.get_length()
        new_size, new_offset = size + cdata_length + SIZE_INT, offset + cdata_length + SIZE_INT

        self.headers[0] = self.file_header = BitmapFileHeader(
            [self.file_header.header[0]] +
            [new_size] +
            self.file_header.header[2:4] +
            [new_offset]
        )
        super().write(data_stream)
        self.headers[0] = self.file_header = BitmapFileHeader(
            [self.file_header.header[0]] +
            [size] +
            self.file_header.header[2:4] +
            [offset]
        )

    @classmethod
    def read_crypto_header(cls, data_stream) -> CryptoHeader:
        """
        Reads bitmap headers from bytestream.
        """
        size = data_stream.read_int()
        data = data_stream.read(size)
        new_header = CryptoHeader([size, data])
        new_header.set_length(size)
        return new_header



"""
    author: Hosung Lee
    date: December 7 2020

    Abstract Cipher class is a base class of all ciphers used within photocrypt package.
"""
from abc import ABC, abstractmethod
from typing import Optional, List


class Cipher(ABC):
    """
    Abstract cipher class
    """

    @abstractmethod
    def encrypt(self, data: bytes, extra: Optional[List[bytes]] = None) -> (bytes, bytes):
        """
        encrypts bytes

            Parameters:
                data (bytes): data to encrypt
                extra (Optional[List[bytes]]): optional data required for encryption

            Return:
                encrypted (bytes): encrypted data
                extra (Optional[bytes]): packed bytes for extra
                                            (packed using photocrypt.pack).
        """
        ...

    @abstractmethod
    def decrypt(self, data: bytes, extra: Optional[List[bytes]] = None) -> (bytes, bytes):
        """
        decrypts bytes

            Parameters:
                data (bytes): data to decrypt
                extra (Optional[List[bytes]]): optional data required for decryption

            Return:
                decrypted (bytes): encrypted data
                extra (Optional[bytes]): packed bytes for extra
                                            (packed using photocrypt.pack).
        """
        ...



"""
Image module. Provides image formats used by photocrypt.
"""

# basic image operations
from photocrypt.core import Image
from photocrypt.core import open_image_as
from photocrypt.core import convert_image
from photocrypt.core.image import _open_image

# import supported image formats
from .bitmap import Bitmap
from .crypto_bitmap import CryptoBitmap

# supported formats
supported_formats = [
    CryptoBitmap,
    Bitmap
]

# open image operation
def open_image(file_path: str) -> Image:
    """
    loads image from path to Image class.
    """

    return _open_image(supported_formats, file_path)



"""
    author: Sean Kullmann
    date: December 7 2020

    provides RSA cipher and generate key feature
"""
from typing import Optional, List, Tuple
from Cryptodome.PublicKey import RSA
from Cryptodome.Cipher import PKCS1_OAEP
from photocrypt.core import Cipher, packer


class RSACipher(Cipher):
    """
    Abstract RSA Cipher class
    """

    def __init__(self, key: bytes):
        self._key = RSA.import_key(key)
        self._rsa = PKCS1_OAEP.new(self._key)

    @property
    def key(self):
        """
        getter method of key
        """
        return self._key

    def encrypt(self, data: bytes, extra: Optional[bytes] = None) -> List[bytes]:
        """
        encrypts bytes

            Parameters:
                data (bytes): data to encrypt using AES
                extra (Optional[bytes]): packed bytes for extra
                                            (packed using photocrypt.pack).

            Return:
                encrypted (bytes): encrypted data
                extra (Optional[bytes]): packed bytes for extra
                                            (packed using photocrypt.pack).
        """
        return self._rsa.encrypt(data), packer.pack(b'')

    def decrypt(self, data: bytes, extra: Optional[bytes] = None) -> List[bytes]:
        """
        decrypts bytes

            Parameters:
                data (bytes): data to decrypt using AES
                extra (Optional[bytes]): packed bytes for extra
                                            (packed using photocrypt.pack).

            Return:
                decrypted (bytes): encrypted data
                extra (Optional[bytes]): packed bytes for extra
                                            (packed using photocrypt.pack).
        """
        return self._rsa.decrypt(data), packer.pack(b'')


def generate_key(key_size: int = 2048) -> (bytes, bytes):
    """
    generates RSA private and public key

            Parameters:
                key_size (int): key size of rsa

            Return:
                private key (bytes): rsa private key
                public key (bytes): rsa public key
    """
    key = RSA.generate(key_size)
    return key.export_key(), key.publickey().export_key()


def save_key(key: bytes, file_path: str) -> None:
    """
    saves single RSA key to given file path.
    Use save_keypair() to save both private and public key at once.

        Parameters:
            key (bytes): byte representation of key (private/public)
            file_path (str): file path to save key.
    """
    with open(file_path, "wb") as file:
        file.write(key)


def load_key(file_path: str) -> bytes:
    """
    loads single RSA key from given file path.
    Use load_keypair() to load both private and public key at once.

    Parameters:
        file_path (str): file path to load key.
    """
    with open(file_path, "rb") as file:
        key = file.read()
    return key


def save_keypair(keypair: Tuple[bytes, bytes], file_paths: Tuple[str, str]) -> None:
    """
    saves RSA keypair to given file path.
    Use save_key() to save a single key.

        Parameters:
            keypair (tuple of bytes): byte representations of keys in tuple (private, public)
            file_path (tuple of str): file paths in pair to save keyspairs.
    """
    if len(keypair) != 2 or len(file_paths) != 2:
        raise ValueError("The tuple has to be size of 2")

    for key, file_path in zip(keypair, file_paths):
        save_key(key, file_path)


def load_keypair(file_paths: Tuple[str, str]) -> Tuple[bytes, bytes]:
    """
    loads RSA keypair from given file path.
    Use load_key() to load a single key.

        Parameters:
            keypair (tuple of bytes): byte representations of keys in tuple (private, public)
            file_path (tuple of str): file paths in pair to save keyspairs.
    """
    if len(file_paths) != 2:
        raise ValueError("The tuple has to be size of 2")

    return (load_key(file_paths[0]), load_key(file_paths[1]))


def create(key: bytes):
    """
    creates RSA Cipher

        Parameters:
            key (bytes): key for cipher
    """
    return RSACipher(key)



"""
    author: Hosung Lee
    date: December 7 2020

    ByteStream class is used for processing bytes data as a stream.
"""

import io
import struct

a = io.BytesIO()
class ByteStream(io.BytesIO):
    """
    ByteStream class extends io.BytesIO

    added read and write method of different types.
    """
    def read_int(self,*, endian: str = 'little') -> int:
        """
        reads int from byte stream.
        """
        return int.from_bytes(self.read(4), endian)

    def read_short(self,*, endian: str = 'little') -> int:
        """
        reads short from byte stream.
        """
        return int.from_bytes(self.read(2), endian)

    def read_str(self, size: int = 1, encoding: str = 'utf-8') -> str:
        """
        reads string from byte stream.
        """
        return self.read(size).decode(encoding)

    def read_type(self, rtype, *args, **kwargs):
        """
        reads of type
        """
        if rtype is bytes:
            return self.read(*args, **kwargs)

        if rtype is int:
            return self.read_int(*args, **kwargs)

        if rtype == 'short':
            return self.read_short(*args, **kwargs)

        if rtype is str:
            return self.read_str(*args, **kwargs)

        if isinstance(rtype, tuple):
            if len(rtype) == 1:
                return self.read_type(rtype[0])

            if len(rtype) == 2:
                return self.read_type(rtype[0], *rtype[1])

            if len(rtype) == 3:
                return self.read_type(rtype[0], *rtype[1], **rtype[2])

        raise TypeError(f"Cannot read this type: {rtype}")

    def read_multiple(self, types: list) -> list:
        """
        reads as specified in the list
        example:
        >>> bs = ByteStream(b'\x01\x00\x00\x00hello\x01\x00')
        >>> bs.read_multiple([int, (str, 5), short])
        [1, "hello", 1]
        """

        return [self.read_type(t) for t in types]

    def write_int(self, data: int, *, endian: str = 'little') -> int:
        """
        writes int to byte stream.
        """
        return self.write(struct.pack(('<' if endian == 'little' else '>') + 'i', data))

    def write_short(self, data: int, *, endian: str = 'little') -> int:
        """
        writes short to byte stream.
        """
        return self.write(struct.pack(('<' if endian == 'little' else '>') + 'h', data))

    def write_str(self, data: str, encoding: str = 'utf-8') -> int:
        """
        writes string to byte stream.
        """
        return self.write(data.encode(encoding))

    def write_type(self, wtype, *args, **kwargs) -> int:
        """
        writes of type
        """
        if wtype is bytes:
            return self.write(*args, **kwargs)

        if wtype is int:
            return self.write_int(*args, **kwargs)

        if wtype == 'short':
            return self.write_short(*args, **kwargs)

        if wtype is str:
            return self.write_str(*args, **kwargs)

        if isinstance(wtype, tuple):
            if len(wtype) == 2:
                return self.write_type(wtype[0], wtype[1])

            if len(wtype) == 3:
                return self.write_type(wtype[0], wtype[1], *wtype[2])

            if len(wtype) == 4:
                return self.write_type(wtype[0], wtype[1], *wtype[2], **wtype[3])

        raise TypeError(f"Cannot write this type: {wtype}")

    def write_multiple(self, types: list) -> None:
        """
        write as specified in the list
        example:
        >>> bs = ByteStream(b'\x01\x00\x00\x00hello\x01\x00')
        >>> bs.write_multiple([(int, 1), (str, "hello"), ('short', 1)])
        [1, "hello", 1]
        """

        for t in types:
            self.write_type(t)



"""
    author: Sean Kullmann, Hosung Lee
    date: December 7 2020

    Photocrypt module to encrypt/decrypt images
"""

from Cryptodome.Random import get_random_bytes
from photocrypt.crypto import AES
from photocrypt.crypto import RSA
from photocrypt.image import CryptoBitmap, Image, convert_image
from photocrypt import packer


def encrypt_image(image: Image, public_key: bytes, *, aes_mode=AES.MODE_EAX) -> Image:
    """
    converts image to CryptoBitmap, then uses AES and RSA to encrypt image.
    """
    # converts image to crypto-bitmap
    if not isinstance(image, CryptoBitmap):
        image = convert_image(image, CryptoBitmap)

    # encrypts image using AES, and encrypts key using RSA
    aes = AES.create(aes_mode, get_random_bytes(16))
    rsa = RSA.create(public_key)
    key_enc, _ = rsa.encrypt(aes.key)
    cdata, extra = aes.encrypt(image.data)
    image.data = cdata
    image.store_crypto_information(key_enc, *packer.unpack(extra))
    return image


def decrypt_image(image: Image, private_key: bytes, *, aes_mode=AES.MODE_EAX) -> Image:
    """
    Decrypts cryptoBitmap using AES and RSA.
    """
    # raise exception if image is not crypto-bitmap
    if not isinstance(image, CryptoBitmap):
        raise TypeError("decryption requires crypto image.")
    image = CryptoBitmap.from_bytes(image.to_bytes())
    unpacked = packer.unpack(image.crypto_header.header[1])
    if len(unpacked) < 3:
        raise ValueError("cannot unpack crypto information.")
    key_dec, nonce, extra = unpacked[0], unpacked[1], unpacked[2]
    rsa = RSA.create(private_key)
    key, _ = rsa.decrypt(key_dec)
    aes_extra, decrypt_extra = packer.pack(nonce), packer.pack(extra)
    aes = AES.create(aes_mode, key, aes_extra)
    pdata, _ = aes.decrypt(image.data, decrypt_extra)
    image.data = pdata
    return image



"""
    author: Hosung Lee
    date: December 7 2020

    packer combines/extracts multiple meaningful data into/from one.
    When packer packs, it prepends size of each data (4 byte integer) to separate the data.
    b'<length1><data1><length2><data2>...<length_n><data_n>'
"""

from typing import List
from .bstream import ByteStream

def pack(*args: bytes) -> bytes:
    """
    prepend four bytes representing number of the data to each data and merge data together.

    >>> pack(b'\x00\x01', b'\x11\x12\x13', b'\xf1')
    
    b'\x02\x00\x00\x00\x00\x01\x03\x00\x00\x00\x11\x12\x13\x01\x00\x00\x00\xf1'
    """
    stream = ByteStream()
    for data in args:
        stream.write_int(len(data))
        stream.write(data)
    return stream.getvalue()

def unpack(packed_data: bytes) -> List[bytes]:
    """
    parse data separated by 4 bytes header of each data. Header represents the length of the data.

    >>> unpack(b'\x02\x00\x00\x00\x00\x01\x03\x00\x00\x00\x11\x12\x13\x01\x00\x00\x00\xf1')


    """
    stream = ByteStream(packed_data)
    result = []
    while stream.tell() < len(packed_data):
        result.append(stream.read(stream.read_int()))
    return result

if __name__ == '__main__':
    print(pack(b'\x00\x01', b'\x11\x12\x13', b'\xf1'))
    print(unpack(b'\x02\x00\x00\x00\x00\x01\x03\x00\x00\x00\x11\x12\x13\x01\x00\x00\x00\xf1'))



"""
setup file for photocrypt package.
"""

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="photocrypt",
    version="0.0.4",
    author="Hosung Lee, Sean Kullmann",
    author_email="runway3237@gmail.com, seankullmann@gmail.com",
    description="A package that can encrypt and decrypt images",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Kullmann/RSAPhotoCryptography",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],
    python_requires='>=3.6',
    keywords='cryptology, cryptograpy, image',
    install_requires=['Pillow','pycryptodome','pycryptodomex'],
    extras_require={
        'gui':['PyQt5', 'PyQt5-sip']
    },
    project_urls={
        'Bug Reports': 'https://github.com/Kullmann/RSAPhotoCryptography/issues',
        'Source': 'https://github.com/Kullmann/RSAPhotoCryptography',
    }

)


"""
Core module. Provides basic operations used by photocrypt.
"""

# iamge operations
from .image import Image
from .image import ImageHeader
from .image import open_image_as
from .image import convert_image

# byte operations
from .bdata import ByteData
from .bstream import ByteStream

# cipher operations
from .cipher import Cipher
